const PUC_CONFIG_NAME = "pucconfig.xml";
const PUC_DEFAULT_CONFIG_NAME = "defaultpucconfig.xml";


// Describes the state of PUC
function pucConfiguration () {

  // checkers hold code that decides how to compare pages
  this.checkers = new Array();

  // pages holds all the information about each page.
  this.pages = new Array();

  // Initialization -- loads the page from disk
  this.Initialize = function() {
    pucCheckPreferences();
    puc_logMessage("Loading Configuration from disk");
    this.loadConfiguration();
  }


  
  // Methods:

  this.toString = function() {
    var retVal = "";
    for(var i = 0; i < this.pages.length; i++)
      retVal += this.pages[i].id + ": " + this.pages[i].title + ", ";
    return retVal;
  }

  this.getPageFromID = function(id) {
    for(var i = 0; i < this.pages.length; i++) 
      if(this.pages[i].id == id)
        return this.pages[i];
    return null;
  }

  this.getPageFromURL = function(url) {
    for(var i = 0; i < this.pages.length; i++) 
      if(this.pages[i].url == url)
        return this.pages[i];
    return null;
  }

  this.getCheckerFromName = function(name) {
    for(var i = 0; i < this.checkers.length; i++) 
      if(this.checkers[i].name == name)
        return this.checkers[i];
    return null;
  }

  // Creates a deep copy of a Configuration, given an initialized one "newConfig"
  this.copyInitialize = function (other) {
    for(var i = 0; i < other.checkers.length; i++) {
      var newChecker = new pucChecker();
      this.checkers.push(newChecker.copyInitialize(other.checkers[i]));
    }

    for(var i = 0; i < other.pages.length; i++) {
      var newPage = new pucPage();
      this.pages.push(newPage.copyInitialize(other.pages[i]));
      
      // These pages will be pointing to the original checkers, which might be ok,
      // but for the options menu, I'd like them to see the newest version,
      // so we have to slightly hack it a level above where we should.
      // Ideally I should have always had a page hold the string of the checker's name,
      // but I like the convenience of being able to go page.checker.blah, not
      // pucConfig.getCheckerFromName(page.checker).blah.
      newPage.checker = this.getCheckerFromName(newPage.checker.name);
    }

    return this;  
  }

  // Loads the configuration file
  this.loadConfiguration = function() {
    var noConfigFile = false;

    // As of Firefox 1.5, the extensions/{PUC_GUID}/store directory is cleared out after each install.
    
    // This version of PUC will store its configuration in the chrome directory.
    // Although it is futile to try to maintain forwards compaitiblity in PUC between,
    // version 0.2 and 0.3 on Firefox 1.5, older versions of firefox will still
    // be able to retail settings, so if we don't see the configuration file in the chrome
    // directory, check the GUID/store directoy.

    try { 
      // First look for the configuration in the chrome directory
      // (this is where it is saved in the newest version of PUC)
      istream = puc_getReadStreamChromeDir(PUC_CONFIG_NAME);
    } catch(e) {
      try {
        // If we didn't find a configuration file there, look for
        // it in the GUID/store directory
        istream = puc_getReadStream(PUC_CONFIG_NAME);
      } catch(e) {
        // There is no existing configuration file
        noConfigFile = true;
      }
    }

    if(!noConfigFile) {
      this.loadStream(istream, true);
      istream.close();
    }
  
    // Read in the default configuration (do so unconditionally because
    // there may be new Checkers in a new version of PUC to be loaded)
    try {
      // The default configuration file is stored in the store directory.
      istream = puc_getReadStream(PUC_DEFAULT_CONFIG_NAME);
    } catch(e) {
      puc_logMessage(e + "\nError: No default configuration file found, cannot initialize Page Update Checker.");
      alert("Page Update Checker configuration corrupted, please reinstall.");
      return;
    }

    this.loadStream(istream, false);
    istream.close();

  }

  // Loads an XML file. addDuplicates == true will include duplicate names of checkers using
  // makeCheckerNameUnique
  this.loadStream = function (istream, addDuplicates) {
    var doc = null;
    var parser = new DOMParser();
    var newItems = {pages: new Array(), checkers: new Array()};
  
    try {
      // This seems to throw an error on my colinux, debian, box:
      doc = parser.parseFromStream(istream, "utf8", istream.available(), "text/xml");
  
      var pXPE = new XPathEvaluator();
      var nodes = pXPE.evaluate("/", doc, null, 0, null);
  
      var configVer = doc.firstChild.getAttribute("version");
      if (configVer > PUC_VERSION) {
        puc_logMessage("Page Update Checker: Config file has more recent version ("+
          configVer+"), than current version of PUC (" + PUC_VERSION + ")");
        puc_logMessage("Error - Version mismatch");
        return null;
      }
  
      if (nodes != null)
      {
        for (var node = null; (node = nodes.iterateNext()); )
        {
          var configs = 0;
          var configNode = null;
  
          for(; (configNode = node.childNodes[configs]); configs++)
          {
            var subNodes = 0;
            var subNode = null;
  
            for(; (subNode = configNode.childNodes[subNodes]); subNodes++)
            {
              if (subNode.nodeName == "page")
              {
                var page = new pucPage();
                page.XMLInitialize(subNode);
                page.id = this.getNewPageId();
                this.pages.push(page);
                newItems.pages.push(page);
              } else if (subNode.nodeName == "checker") {
                var checker = new pucChecker();
                checker.XMLInitialize(subNode);

                if(addDuplicates) {
                  this.makeCheckerNameUnique(checker);
                  this.checkers.push(checker);
                  newItems.checkers.push(checker);
                } else {
                  // Make sure that there isn't already a checker with this name
                  var isUnique = true;
                  for(var i = 0; i < this.checkers.length; i++) {
                    if(this.checkers[i].name == checker.name) {
                      isUnique = false;
                      break;
                    }
                  }
                  if(isUnique) {
                    this.checkers.push(checker);
                    newItems.checkers.push(checker);
                  }
                }
              }
            }
          }
        }
      }
      // Assign the appropriate checker to each page
      for(var i = 0; i < this.pages.length; i++) {
        this.pages[i].checker = this.getCheckerFromName(this.pages[i].checkerXMLPointer);

        if(this.pages[i].checker == null) {
          // Either the page didn't have a checker assigned, or the checker didn't exist.
          // Just assign the first checker, if there is one.
          //
          // In this way, when people upgrade, and so long as the first default config is
          // "Basic", they won't see any unexpected change in behavior.
          if(this.checkers[0])
            this.pages[i].checker = this.checkers[0];
        }
      }
    }
    catch (e)
    {
      puc_logMessage(e + "\nThis probably means that you will not be able to save your settings when you close firefox.");
    }
    return newItems;
  }

  // This function saves a page in the XML config file
  this.saveConfiguration = function(){  
    try
    {
      var doc = document.implementation.createDocument("", "pucconfig", null);
  
      for (var i = 0; i < this.checkers.length; i++)
      {
        var checkerObj = this.checkers[i];
        if (checkerObj != null) {
          var checkerNode = checkerObj.toXMLNode(doc);
  
          doc.firstChild.appendChild(doc.createTextNode("\n\t"));
          doc.firstChild.appendChild(checkerNode);
        }
      }

      for (var i = 0; i < this.pages.length; i++)
      {
        var pageObj = this.pages[i];
        if (pageObj != null) {
          var pageNode = pageObj.toXMLNode(doc, true);
  
          doc.firstChild.appendChild(doc.createTextNode("\n\t"));
          doc.firstChild.appendChild(pageNode);
        }
      }
  
      doc.firstChild.appendChild(doc.createTextNode("\n"));
      doc.firstChild.setAttribute("version", PUC_VERSION);
  
      // In PUC 0.2 and down, we saved the configuration file in the 
      // {GUID}/store directory, but now that Firefox 1.5 deletes these 
      // files, we've moved it to the chrome directory.
      var configStream = puc_getWriteStreamChromeDir(PUC_CONFIG_NAME);
      new XMLSerializer().serializeToStream(doc, configStream, "utf-8");
      configStream.close();
    }
    catch (e)
    {
       puc_logMessage("Error while trying to save PUC configuration: " + e);
    }
  }

  this.getNewPageId = function() {
    // Find the highest available ID, avoiding conflicts
    var newId = 0;
    for(var i = 0; i < this.pages.length; i++)
      if(this.pages[i].id * 1 >= i)      // We multiply by one to force it to treat id as a number
        newId = this.pages[i].id * 1 + 1;
    return newId;

}

  this.addPage = function(newTitle, newUrl) {
    pucNewPage = new pucPage();

    with(pucNewPage) {
      enabled = true;
      id = this.getNewPageId();

      title = newTitle;
      url = newUrl;
      // The first checker will be the default
      if(this.checkers[0])
        checker = this.checkers[0];

      puc_logMessage("New Page being added with ID " + id);
    }
    this.pages.push(pucNewPage);

    return pucNewPage;
  }

  // removePage returns the removed page
  this.removePage = function(page) {
    for(var i = 0; i < this.pages.length; i++) {
      if(this.pages[i] == page) {
        this.pages.splice(i, 1);
        return page;
      }
    }
    return null;
  }

  // Add a new checker, (the name must be unique)
  this.addChecker = function(newChecker) {
    // Make sure no other checkers have this name, change it if necessary
    this.makeCheckerNameUnique(newChecker);

    puc_logMessage("New Checker being added " + newChecker.name);
    this.checkers.push(newChecker);
    return newChecker;
  }

  this.makeCheckerNameUnique = function(checker) {
    for(var i = 0; i < this.checkers.length; i++) {
      if(this.checkers[i] != checker && this.checkers[i].name == checker.name) {
        checker.name += "-new";
        puc_logMessage("Name of new: " + checker.name);
        // Recurse until we don't have duplicate names
        this.makeCheckerNameUnique(checker);
        return;
      }    
    }
  }

  // removeChecker returns the removed checker
  this.removeChecker = function(checker) {
    for(var i = 0; i < this.checkers.length; i++) {
      if(this.checkers[i] == checker) {
        this.checkers.splice(i, 1);
        return checker;
      }
    }
    return null;
  }

}



// Makes sure that preferences are loaded, with defaults
function pucCheckPreferences()
{
      pucSetDefaultPref("puc.notifydespiteupdate", Components.interfaces.nsIPrefBranch.PREF_BOOL, false);
      pucSetDefaultPref("puc.alertonfailure", Components.interfaces.nsIPrefBranch.PREF_BOOL, true);
      pucSetDefaultPref("puc.notifymethod", Components.interfaces.nsIPrefBranch.PREF_STRING, "dialog");
      pucSetDefaultPref("puc.playsound", Components.interfaces.nsIPrefBranch.PREF_BOOL, false);
      pucSetDefaultPref("puc.soundlocation", Components.interfaces.nsIPrefBranch.PREF_STRING, "chrome://puc/skin/notify.wav");
      pucSetDefaultPref("puc.showactivity", Components.interfaces.nsIPrefBranch.PREF_BOOL, true);
}

// If we don't have any defaults set then create them with the appropriate values!
function pucSetDefaultPref(prefName, prefType, value)
{
      var prefToSet = PrefUtils.getPref(prefName);
      if (prefToSet == null)
      {
          PrefUtils.setPrefWithType(prefName, prefType, value);
      }
}